package de.zuim.ledcontrol.effects;

import com.tagtraum.jipes.math.FFTFactory;
import com.tagtraum.jipes.math.Transform;
import de.zuim.ledcontrol.LEDControl;
import de.zuim.ledcontrol.LEDEffect;

import javax.sound.sampled.*;
import java.awt.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static de.zuim.ledcontrol.LEDControl.HEIGHT;
import static de.zuim.ledcontrol.LEDControl.WIDTH;

public class AudioEffect implements LEDEffect {
    @Override
    public String getDescription() {
        return "Audio Effect";
    }

    final static int BUFFERSIZE = 2048;
    double[] magnitudes = null;
    Thread audioThread = null;

    @Override
    public void load() {
        Runnable audioRunnable = () -> {


            Mixer.Info[] mixers = AudioSystem.getMixerInfo();
            List<Line.Info> availableTargetLines = new ArrayList<>();
            for (Mixer.Info mixerInfo : mixers){

                Mixer m = AudioSystem.getMixer(mixerInfo);

                Line.Info[] lines = m.getTargetLineInfo();

                for (Line.Info li : lines)
                {
                    try
                    {
                        if(li instanceof DataLine.Info && mixerInfo.toString().contains("mix"))
                        {
                            m.open();
                            System.out.println("("+availableTargetLines.size()+") Found target line: " + li+"  "+mixerInfo + "("+li.getClass()+")");
                            availableTargetLines.add(li);
                            m.close();
                        }
                    } catch (LineUnavailableException e){
                        System.out.println("Line unavailable.");
                    }
                }
            }

            DataLine.Info targetLineInfo = (DataLine.Info) availableTargetLines.get(0);

            System.out.println("SUPPORTED TARGET FORMATS: ");
            AudioFormat[] formats = (targetLineInfo).getFormats();
            for(int i=0; i<formats.length;i++)
            {
                System.out.println("("+i+")"+formats[i]);
            }
            AudioFormat format = formats[2];


            System.out.println("SELECTED: "+format);


            final DataLine.Info info = new DataLine.Info(TargetDataLine.class, format);
            final AudioInputStream audioStream;

            try  {
                TargetDataLine targetLine = (TargetDataLine) AudioSystem.getLine(info);
                targetLine.open();
                targetLine.start();
                audioStream = new AudioInputStream(targetLine);

                final byte[] buf = new byte[BUFFERSIZE];

                final int numberOfSamples = buf.length / format.getFrameSize();
                final Transform fft = FFTFactory.getInstance().create(numberOfSamples);
                while (audioThread != null) {
                    // in real impl, don't just ignore how many bytes you read
                    int read = audioStream.read(buf);
                    // the stream represents each sample as two bytes -> decode

                    float[] fbuf = decode(buf,format);

                    final float[][] transformed = fft.transform(fbuf);
                    final float[] realPart = transformed[0];
                    final float[] imaginaryPart = transformed[1];
                    magnitudes = toMagnitudes(realPart, imaginaryPart);

                    //System.out.println("M"+ Arrays.toString(magnitudes));
                    // do something with magnitudes...
                    int max = 0, avg = 0,min=11111111;
                    for(double m : magnitudes){
                        avg += m;
                        if(m>max)
                            max= (int) m;
                        if(m<min)
                            min= (int) m;
                    }
                    //System.out.println("M"+ max+"  "+min+"   "+avg/magnitudes.length+ "  "+read+"  "+magnitudes.length);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        };

        audioThread = new Thread(audioRunnable);
        audioThread.start();
    }

    @Override
    public void unload() {
        if(audioThread != null && audioThread.isAlive())
            audioThread = null;
    }

    @Override
    public void render(long timeDelta, Graphics g) {
        for(int x = 0; x< WIDTH; x++){
            double val = 0;
            if(magnitudes != null) {
                for (int sample = 0; sample < magnitudes.length / WIDTH; sample++) {
                    val += magnitudes[x * (magnitudes.length / WIDTH) + sample];
                }
                val /= (double) magnitudes.length / WIDTH;
            }
            int y= (int) Math.round(Math.max(0,val-1)/10);
            g.setColor(new Color(Math.abs(y)+1,2,0));
            g.drawRect(x,0,1,y);
        }
    }



    private static final float NORMALIZATION_FACTOR_2_BYTES = Short.MAX_VALUE + 1.0f;
    private static float[] decode(final byte[] buf, final AudioFormat format) {
        final float[] fbuf = new float[buf.length / format.getFrameSize()];
        for (int pos = 0; pos < buf.length; pos += format.getFrameSize()) {
            final int sample = format.isBigEndian()
                    ? byteToIntBigEndian(buf, pos, format.getFrameSize())
                    : byteToIntLittleEndian(buf, pos, format.getFrameSize());
            // normalize to [0,1] (not strictly necessary, but makes things easier)
            fbuf[pos / format.getFrameSize()] = sample / NORMALIZATION_FACTOR_2_BYTES;
        }
        return fbuf;
    }

    private static double[] toMagnitudes(final float[] realPart, final float[] imaginaryPart) {
        final double[] powers = new double[realPart.length / 2];
        for (int i = 0; i < powers.length; i++) {
            powers[i] = Math.sqrt(realPart[i] * realPart[i] + imaginaryPart[i] * imaginaryPart[i]);
        }
        return powers;
    }

    private static int byteToIntLittleEndian(final byte[] buf, final int offset, final int bytesPerSample) {
        int sample = 0;
        for (int byteIndex = 0; byteIndex < bytesPerSample; byteIndex++) {
            final int aByte = buf[offset + byteIndex] & 0xff;
            sample += aByte << 8 * (byteIndex);
        }
        return sample;
    }

    private static int byteToIntBigEndian(final byte[] buf, final int offset, final int bytesPerSample) {
        int sample = 0;
        for (int byteIndex = 0; byteIndex < bytesPerSample; byteIndex++) {
            final int aByte = buf[offset + byteIndex] & 0xff;
            sample += aByte << (8 * (bytesPerSample - byteIndex - 1));
        }
        return sample;
    }
}
